/* $Id: loader.S,v 1.11 2004/03/11 02:40:05 sei_core Exp $ */

/*
 * Boot Loader for Sony BSNC DIF-144, PDW-V1/5x0
 * 
 * Copyright 2002, 2003, 2004 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 *
 */

/* #define JIG_SEL BSNC_DIF144_DIPSW_JIG_BOOT_OLD */
#define JIG_SEL BSNC_DIF144_DIPSW_JIG_BOOT

/*
 * NSC Linux kernel headers.
 */
#include <asm-mips/asm.h>
#include <asm-mips/regdef.h>
#include <asm-mips/mipsregs.h>
#include <asm-mips/vr4122/vr4122.h>
#include <asm-mips/vr41xx/dif144.h>
#include "config.h"

#define SLEEP(n) li s0, n; 10: bnez s0, 10b; sub s0, 1

#undef LEDDEBUG /* Do NOT define this on dvp27. */
#ifdef LEDDEBUG
#define LED(n) li a0, n; jal debug_led; nop;
#else
#define LED(n)
#endif

#define NEW_BOOT_MODE
#define LNX_INT_X	(1<<6)
#define PPCI_INTA_X	(1<<4)

    .set mips3

LEAF(_start)
    .set    noreorder

    LED(0x01)

    jal	    clkset_and_reset
    nop

    /* Initialize hardware (BCU,ROM,RAM) */
    jal	    init_hw
    nop

    /* Enable cache */
    LED(0x02)
    mfc0    t0, CP0_CONFIG
    nop
    ori	    t0, 0x2
    xori    t0, 0x2
    mtc0    t0, CP0_CONFIG
    nop

    /* Initialize cache */
    jal	    init_cache
    nop

    li	    t0, 0x9fc00040  /* jump to kseg0 */
    jr	    t0
    nop

    nop
    nop
    nop
    nop
    nop

    /*
     * kernel select (A: normal kernel,  B: rescue kernel)
     *
     * How to select
     * =============
     *
     * <BSNC_DIF144_DIPSW_SEL_BY_DIPSW [1]>
     *     |
     *     |--- on(0) ----+
     *     |              |
     *   off(1)           V
     *     |        <BSNC_DIF144_DIPSW_SEL_KERNEL? [4]> --off(1)--> kernel A
     *     V                             |                             |
     * <ElapsedTimer reset?[2]>          `----on(0)--> kernel B        |
     *     |                                                 `---------+
     *     |--- yes ---> kernel B + rootfs B                           |
     *     |             (normal mode seems to be broken)              |
     *     |                                                           |
     *     no                                                          V
     *     |                                     BSNC_DIF144_DIPSW_SEL_ROOTFS
     *     V                                             |             |
     * <which bootmode of itron? [3]>                    |             |
     *     |                                           off (1)       on (0)
     *     |--- hibernate                                |             |
     *     |                                          rootfs A      rootfs B
     *     |--- normal --> kernel A + rootfs A            (flash or jig)
     *     |                     (flash or jig) 
     *     `--- recovery --> kernel B + rootfs B
     *                           (flash or jig)
     */

    /* [1] check for BSNC_DIF144_DIPSW_SEL_BY_DIPSW */
    lh      t1, BSNC_DIF144_DIPSW
    andi    t0, t1, BSNC_DIF144_DIPSW_SEL_BY_DIPSW
    beq     t0, zero, sel_kernel_by_dipsw
    nop

    /* [2] check for ElapsedTimer reset */
    lh      t0, VR4122_RTCINTREG
    andi    t1, t0, 0x1
    bgtz    t1, force_rescue
    nop

#ifndef NEW_BOOT_MODE
    /* [3] check for selector on flash */
/* sel_by_from: */
    li	    t0, 0xbffe4004 /* TODO remove magic number */
    lw	    t1, (t0)
    and     t0, t1, 0x00010000
    bgtz    t0, kernel_a
    nop
    j       kernel_b
    nop
#else
    /* [3] check for itron boot mode */
/* sel_by_itron: */
waiting:
    SLEEP(200) 
    nop
    lh      t0, VR4122_GIUPIODL
    andi    t1, t0, LNX_INT_X
    beq     t1, zero, skip_wait
    nop
    andi    t1, t0, PPCI_INTA_X
    beq     t1, zero, skip_wait
    nop
    j       waiting
    nop
skip_wait:
    SLEEP(2000000) 
    nop
    /* hibernate? */
    lh      t0, VR4122_GIUPIODL
    andi    t1, t0, (LNX_INT_X | PPCI_INTA_X)
    beq	    t1, zero, do_hibernate
    nop
    /* normal? */
    andi    t1, t0, LNX_INT_X 
    beq     t1, zero, kernel_a
    nop
    /* recovery */
    j       force_rescue
    nop
#endif
    /* [4] check for dip sw */
sel_kernel_by_dipsw:
    jal     clear_haltimer
    nop
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, BSNC_DIF144_DIPSW_SEL_KERNEL
    bgtz    t1, kernel_a
    nop
    j       kernel_b
    nop

force_rescue:
    jal     clear_haltimer
    nop

kernel_b:
    li	    s0, KSTART_ADDR_B
    li	    s2, KEND_ADDR_B
    /* flash or jig ? */
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, JIG_SEL
    beq     t1, zero, jig_rootfs_b
    nop
/* rootfs_b: */
    la	    a1, argv_mtd2
    j       done_sel
    nop
jig_rootfs_b:
    la	    a1, argv_mtd9
    j       done_sel
    nop

kernel_a:
    /* set Elapsed Timer */
    jal     set_elapsedtimer
    nop
    li	    s0, KSTART_ADDR_A
    li	    s2, KEND_ADDR_A
    /* flash or jig ? */
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, JIG_SEL
    beq     t1, zero, jig_rootfs_a
    nop
/* rootfs_a: */
    la	    a1, argv_mtd1
    j       done_sel
    nop
jig_rootfs_a:
    la	    a1, argv_mtd8

done_sel:
    li	    s1, KDEST_ADDR

/* recheck BSNC_DIF144_DIPSW for BSNC_DIF144_DIPSW_SEL_BY_DIPSW */
    lh      t1, BSNC_DIF144_DIPSW
    andi    t0, t1, BSNC_DIF144_DIPSW_SEL_BY_DIPSW
    beq     t0, zero, sel_rootfs_by_dipsw
    nop
    j       done
    nop

sel_rootfs_by_dipsw:
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, BSNC_DIF144_DIPSW_SEL_ROOTFS
    bgtz    t1, rootfs_a
    nop

/* rootfs_b: */
    /* flash or jig ? */
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, JIG_SEL
    beq     t1, zero, jig_rootfs_b_by_dipsw
    nop
/* rootfs_b: */
    la	    a1, argv_mtd2
    j       done
    nop
jig_rootfs_b_by_dipsw:
    la	    a1, argv_mtd9
    j       done
    nop

rootfs_a:
    lh      t0, BSNC_DIF144_DIPSW
    andi    t1, t0, JIG_SEL
    beq     t1, zero, jig_rootfs_a_by_dipsw
    nop
    la	    a1, argv_mtd1
    j       done
    nop
jig_rootfs_a_by_dipsw:
    la	    a1, argv_mtd8

done:

    LED(0x04)

    /* Copy kernel binary to SDRAM from FlashROM */
1:
    lw	    t0, (s0)
    sw	    t0, (s1)

    add	    s0, 4
    add	    s1, 4

    bne	    s0, s2, 1b
    nop

    LED(0x08)

    lw	    a0, argc

    li	    t0, ENTRY_POINT
    jr	    t0

    nop

END(_start)

LEAF(debug_led)
    .set    noreorder

    la      t0, 0xaa000000
    xori    a0, 0xff
    sb      a0, 0(t0)

    jr	    ra
    nop
END(debug_led)

LEAF(init_hw)
    .set    noreorder

    /* config register */
    mtc0    zero, CP0_WATCHLO
    nop
    mtc0    zero, CP0_WATCHHI
    nop
    mtc0    zero, CP0_STATUS
    nop

    /* BCU registers */
    li	    a1, BCU_CNT1
    sh	    a1, VR4122_BCUCNTREG1
    li	    a1, BCU_CNT3
    sh	    a1, VR4122_BCUCNTREG3

    /* ROM */
    li	    a1, ROM_SPEED
    sh	    a1, VR4122_ROMSPEEDREG
    li	    a1, ROM_SIZE
    sh	    a1, VR4122_ROMSIZEREG

    /* external I/O */
    li	    a1, IO0_SPEED
    sh	    a1, VR4122_IO0SPEEDREG
    li	    a1, IO1_SPEED
    sh	    a1, VR4122_IO1SPEEDREG

    /* SDRAM */
    li	    a1, RAM_SIZE
    sh	    a1, VR4122_RAMSIZEREG
    li	    a1, BCU_RFCNT
    sh	    a1, VR4122_BCURFCNTREG
    li	    a1, SDRAM_MODE
    sh	    a1, VR4122_SDRAMMODEREG
    sh	    a1, VR4122_SDRAMMODEREG
    li	    a1, SDRAM_CNT
    sh	    a1, VR4122_SDRAMCNTREG

    /*
     * disable interrupt
     */
    mtc0    zero,   CP0_STATUS
    nop
    /* level 1 */
    sh	    zero,   VR4122_MSYSINT1REG
    sh	    zero,   VR4122_MSYSINT2REG
    /* level 2 */
    sh	    zero,   VR4122_SOFTINTREG
#if 0
    sh	    zero,   VR4122_MBCUINTREG
#endif
    sh	    zero,   VR4122_MFIRINTREG
    sh	    zero,   VR4122_MDSIUINTREG
    sh	    zero,   VR4122_MGIUINTLREG
    sh	    zero,   VR4122_MGIUINTHREG
    sh	    zero,   VR4122_MCSIINTREG
    sh	    zero,   VR4122_MSCUINTREG
    sh	    zero,   VR4122_MPCIINTREG

    /*
     * GPIO setting
     *
     * GPIO0-soft reset    I/O out
     * GPIO1
     * GPIO2-push switch1
     * GPIO3-push switch2
     * GPIO4-hot line 1    I/O in
     * GPIO5-P2P
     * GPIO6-hot line 2    I/O in
     * GPIO7-LED enable    I/O out
     * GPIO8-CardBus       Act-low
     * GPIO9-PIER	   Act-low?
     * GPIO10-
     * GPIO11-
     * GPIO12-
     * GPIO13-OSD
     */
    /* Init TP301-308, LED */
    li	    a1, BSNC_DIF144_IO_DEFAULT
    sh	    a1, BSNC_DIF144_IO
    /* level settings */
    li	    a1, 0x0 /* 1: hi, 0:low */
    sh	    a1, VR4122_GIUINTALSELL
    /* I/O select */
    li	    a1, 0x0081 /* 0, 7 */
    sh      a1, VR4122_GIUIOSELL
    li	    a1, 0x0001 /* enable LED */
    sh      a1, VR4122_GIUPIODL

#if 0
    li	    a1, 0x1 /* 1:edge 0:level */
    sh	    a1, VR4122_GIUINTTYPL
#endif

    li	    a1, 0x0
    sh	    a1, VR4122_GIUINTENL
    li	    a1, 0xbfff
    sh	    a1, VR4122_GIUINTSTATL

    jr	ra
    nop

END(init_hw)

LEAF(init_cache)
/*
 * I-cache (4122:32KB  4131:16KB(8KB x 2way))
 * D-cache (4122:16KB  4131:16KB(8KB x 2way))
 */
    .set    noreorder

    mtc0    zero, CP0_TAGLO
    nop

    /* i cache */
    li	    t1, 0x2000 /* 8KB */
    li	    t0, 0x80000000
10:
    cache   (2 << 2) | 0, (t0)          /* way 0 */
    cache   (2 << 2) | 0, 0x2000 (t0)   /* way 1 */
    addiu   t1, t1, -16
    bne	    t1, zero, 10b
    addiu   t0, t0, 16


    /* d cache */
    li	    t1, 0x2000
    li	    t0, 0x80000000
20:
    cache   (2 << 2) | 1, (t0)         /* way 0 */                   
    cache   (2 << 2) | 1, 0x2000 (t0)  /* way 1 */
    addiu   t1, t1, -16
    bne	    t1, zero, 20b
    addiu   t0, t0, 16

    jr	    ra
    nop
END(init_cache)

LEAF(clkset_and_reset)
    .set	noreorder

    /*
     * Checking for what cause, boot or reset?
     */
    lh      a0, VR4122_PMUINTREG
    li	    a1, 0xffff
    sh	    a1, VR4122_PMUINTREG
    andi    a1, a0, 0x0010

    /* If this routine called as first, set div. if not do nothing */
    bne     a1, zero, set_div
    nop
    li      a0, 0x0010
    jr	    ra
    nop

set_div:
     /* set VTCLK/TCLK div mode */
#if 1
    /*
     *  VTClock = 100Mhz (PClock  / 2)
     *  TClock  =  25Mhz (VTClock / 4)
     */
    li	    a2, 0x0102
    sh	    a2, VR4122_PMUTCLKDIVREG
#else
    /*
     *  VTClock =  33Mhz (PClock  / 6)
     *  TClock  =  16Mhz (VTClock / 2)
     */
    li	    a2, 0x0006
    sh	    a2, VR4122_PMUTCLKDIVREG
#endif
    /* reset */
    SLEEP(50000)
    li	    a1, 0x0011
    sh	    a1, VR4122_PMUCNT2REG
    LED(0x03)
1:	
    b	    1b
    nop
END(clkset_and_reset)

LEAF(set_elapsedtimer)
    /* clear ElapsedTimer */
    sh	  zero, VR4122_ETIMELREG
    sh	  zero, VR4122_ETIMEMREG
    sh    zero, VR4122_ETIMEHREG
    /* set compare register  (0000 0005 0000 = 10 sec) */
    sh    zero, VR4122_ECMPLREG
    li	    a0, 0x5
    sh      a0, VR4122_ECMPMREG
    sh    zero, VR4122_ECMPHREG
    /* enable ElapsedTimer interrupt */
    li	    a0, 0x8
    sh	    a0, VR4122_MSYSINT1REG

    jr	    ra
    nop
END(set_elapsedtimer)

LEAF(do_hibernate)
    .set	noreorder
    /* clock off (bit2,5,5) */
    li	    a1, BSNC_DIF144_IO_DEFAULT & ~(0x0034)
    sh	    a1, BSNC_DIF144_IO
    nop
    nop
    nop
    nop
    nop
    .word	0x42000023
    nop
    nop
    nop
    nop
    nop
    .set	reorder
END(do_hibernate)

LEAF(clear_haltimer)
    /* haltimer clear */
    lh	    t0, VR4122_PMUCNTREG
    or      t0, 0x4
    sh	    t0, VR4122_PMUCNTREG 
    /* disable ElapsedTimer interrupt */
    sh    zero, VR4122_MSYSINT1REG

    jr	    ra
    nop
END(clear_haltimer)

/*
 * kernel parameters
 *  argv, argc   : normal mode
 *  rargv, rargc : rescue mode
 */
    .align 2
argc:
    .long  2
    .align 2
argv0:
    .string "unused" /* see prom_init() in kernel */
    .align 2

    /* mtd1 (rootfs-a on flashrom) */
argv1_mtd1:
    .string "root=/dev/mtdblock1 ramdisk=16384 ro"
    .align 2
argv_mtd1:
    .long  argv0
    .long  argv1_mtd1
    .long  0
    .align 2

    /* mtd2 (rootfs-b on flashrom) */
argv1_mtd2:
    .string "root=/dev/mtdblock2 ramdisk=32768 ro"
    .align 2
argv_mtd2:
    .long  argv0
    .long  argv1_mtd2
    .long  0
    .align 2

    /* mtd8 (rootfs-a on jig) */
argv1_mtd8:
    .string "root=/dev/mtdblock8 ramdisk=16384 ro"
    .align 2
argv_mtd8:
    .long  argv0
    .long  argv1_mtd8
    .long  0
    .align 2

     /* mtd9 (rootfs-b on jig) */
argv1_mtd9:
    .string "root=/dev/mtdblock9 ramdisk=32768 ro"
    .align 2
argv_mtd9:
    .long  argv0
    .long  argv1_mtd9
    .long  0

